
/**
 * @file
 * @author  Aapo Kyrola <akyrola@cs.cmu.edu>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2012] [Aapo Kyrola, Guy Blelloch, Carlos Guestrin / Carnegie Mellon University]
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 
 *
 * @section DESCRIPTION
 *
 * Template for GraphChi applications. To create a new application, duplicate
 * this template.
 */



#include <string>
#include <fstream>
#include <stdlib.h>

#define GRAPHCHI_DISABLE_COMPRESSION

#include "util/toplist.hpp"
#include "graphchi_basic_includes.hpp"

using namespace graphchi;

/**
  * Type definitions. Remember to create suitable graph shards using the
  * Sharder-program. 
  */
typedef float VertexDataType;
typedef float EdgeDataType;

// /**
//   * GraphChi programs need to subclass GraphChiProgram<vertex-type, edge-type> 
//   * class. The main logic is usually in the update function.
//   */
// struct MyGraphChiProgram : public GraphChiProgram<VertexDataType, EdgeDataType> {
    
 
//     /**
//      *  Vertex update function.
//      */
//     void update(graphchi_vertex<VertexDataType, EdgeDataType> &vertex, graphchi_context &gcontext) {

//         if (ginfo.iteration == 0) {
//             /* On first iteration, initialize vertex (and its edges). This is usually required, because
//                on each run, GraphChi will modify the data files. To start from scratch, it is easiest
//                do initialize the program in code. Alternatively, you can keep a copy of initial data files. */
//             // vertex.set_data(init_value);
        
//         } else {
//             /* Do computation */ 

//             /* Loop over in-edges (example) */
//             for(int i=0; i < vertex.num_inedges(); i++) {
//                 // Do something
//             //    value += vertex.inedge(i).get_data();
//             }
            
//              Loop over out-edges (example) 
//             for(int i=0; i < vertex.num_outedges(); i++) {
//                 // Do something
//                 // vertex.outedge(i).set_data(x)
//             }
            
//             /* Loop over all edges (ignore direction) */
//             for(int i=0; i < vertex.num_edges(); i++) {
//                 // vertex.edge(i).get_data() 
//             }
            
//             // v.set_data(new_value);
//         }
//     }
    
//     /**
//      * Called before an iteration starts.
//      */
//     void before_iteration(int iteration, graphchi_context &gcontext) {
//     }
    
//     /**
//      * Called after an iteration has finished.
//      */
//     void after_iteration(int iteration, graphchi_context &gcontext) {
//     }
    
//     /**
//      * Called before an execution interval is started.
//      */
//     void before_exec_interval(vid_t window_st, vid_t window_en, graphchi_context &gcontext) {        
//     }
    
//     /**
//      * Called after an execution interval has finished.
//      */
//     void after_exec_interval(vid_t window_st, vid_t window_en, graphchi_context &gcontext) {        
//     }
    
// };


/**
  * Faster version of pagerank which holds vertices in memory. Used only if the number
  * of vertices is small enough.
  */
struct RumorPropInmem : public GraphChiProgram<VertexDataType, EdgeDataType> {
    
    std::vector<int> rp;
    RumorPropInmem(int nvertices) :   rp(nvertices, 0) {}
    
    void update(graphchi_vertex<VertexDataType, EdgeDataType> &v, graphchi_context &ginfo) {
        if (ginfo.iteration > 0 && rp[v.id()]==1) {
            for(int i=0; i < v.num_outedges(); i++) {
                if(rp[v.outedge(i)->vertexid]==0) {
                    if((float)0.1<(float)(rand()/RAND_MAX)) {
                        rp[v.outedge(i)->vertexid] = 1;
                    } else {
                        rp[v.outedge(i)->vertexid] = 2;
                    }
                }
            }
        }
    }
    
};

int main(int argc, const char ** argv) {
    /* GraphChi initialization will read the command line 
       arguments and the configuration file. */
    graphchi_init(argc, argv);
    
    /* Metrics object for keeping track of performance counters
       and other information. Currently required. */
    metrics m("rumor-propagation");
    global_logger().set_log_level(LOG_DEBUG);
    
    /* Basic arguments for application */
    std::string filename = get_option_string("file");  // Base filename
    int niters           = get_option_int("niters", 10); // Number of iterations
    bool scheduler       = get_option_int("scheduler", 0); // Whether to use selective scheduling
    
    /* Detect the number of shards or preprocess an input to create them */
    int nshards          = convert_if_notexists<EdgeDataType>(filename, 
                                                            get_option_string("nshards", "auto"));
    
    /* Run */
    graphchi_engine<VertexDataType, EdgeDataType> engine(filename, nshards, scheduler, m);
    engine.set_modifies_inedges(false); // Improves I/O performance.

    //bool inmemmode = engine.num_vertices() * sizeof(EdgeDataType) < (size_t)engine.get_membudget_mb() * 1024L * 1024L;
    //if (inmemmode) {
    logstream(LOG_INFO) << "Running Rumor-Propagation by holding vertices in-memory mode!" << std::endl;
    engine.set_modifies_outedges(false);
    engine.set_disable_outedges(true);
    engine.set_only_adjacency(true);
    RumorPropInmem program(engine.num_vertices());
    engine.run(program, niters);
    //} else {
    //    PagerankProgram program;
    //    engine.run(program, niters);
    //}
     
    /* Report execution metrics */
    metrics_report(m);
    return 0;
}
