/*
 * randomwalk.h
 * LLAMA Graph Analytics
 *
 * Copyright 2014
 *      The President and Fellows of Harvard College.
 *
 * Copyright 2014
 *      Oracle Labs.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#ifndef LL_RANDOMWALK_H
#define LL_RANDOMWALK_H

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <float.h>
#include <limits.h>
#include <cmath>
#include <algorithm>
#include <omp.h>
#include <vector>
#include <queue>

#include "llama/ll_writable_graph.h"
#include "benchmarks/benchmark.h"
#include "llama/ll_utils.h"

using namespace std;

//==========================================================================//
// Utilities                                                                //
//==========================================================================//

struct node_and_data {
	node_t node;
	double data;
};


/**
 * Compare by the node data
 */
int cmp_by_data_desc(const void* a, const void* b) {

	const node_and_data* A = (const node_and_data*) a;
	const node_and_data* B = (const node_and_data*) b;

	if (A->data > B->data) return -1;
	if (A->data < B->data) return  1;
	return (int) A->node - (int) B->node;
}


/**
 * The randomwalk benchmark
 */
template <class Graph, typename value_t>
class ll_b_randomwalk_pull_ext : public ll_benchmark<Graph> {

	value_t d;
	int32_t max;

	value_t* G_pg_rank;

public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_pull_ext(int32_t max, value_t d=0.85)
		: ll_benchmark<Graph>(
				/* assuming that value_t is either a float or a double */
				sizeof(value_t) == sizeof(float)
					? "rumorp<float> - Pull"
					: "rumorp<double> - Pull") {
		assert(d > 0 && d < 1);

        printf("max_iteration<%d>\n", max);
		this->d = d;
		this->max = max;
        printf("max_iteration<%d>\n", this->max);

		this->create_auto_array_for_nodes(G_pg_rank);
	}


	/**
	 * Destroy the benchmark
	 */
	virtual ~ll_b_randomwalk_pull_ext(void) {
	}


	/**
	 * Run the benchmark
	 *
	 * @return the numerical result, if applicable
	 */
	virtual double run(void) {
		printf("Rumor Propagation\n");

		Graph& G = *this->_graph;
		ll_memory_helper m;

		// this->progress_init(max);

#pragma omp parallel
		{

#pragma omp for schedule(dynamic, 4096)
			for(int32_t iter=0; iter<this->max; iter++){

				queue<node_t> q; q.push(ll_rand64_positive()%G.max_nodes());
				unsigned char* G_inf_bitmap = m.allocate<unsigned char>((G.max_nodes()+7)/8);
				#pragma omp for nowait
				for (node_t t0 = 0; t0 < (G.max_nodes()+7)/8; t0 ++){
					G_inf_bitmap[t0] = 0;
				}

				while(!q.empty()){
					const node_t src = q.front(); q.pop();

					ll_foreach_out(w, G, src){
						if(_ll_get_bit(G_inf_bitmap, w)==1) { continue; }
            if((value_t)0.1 < (value_t)rand()/RAND_MAX){
							q.push(w);
						}
						_ll_set_bit(G_inf_bitmap, w);
					}

				}

			}
		}

		return 0;
	}


	/**
	 * Finalize the benchmark
	 *
	 * @return the updated numerical result, if applicable
	 */
	virtual double finalize(void) {
		return 73.0; // dummy value
	}


	/**
	 * Print the results
	 *
	 * @param f the output file
	 */
	virtual void print_results(FILE* f) {
		print_results_part(f, this->_graph, G_pg_rank);
	}
};


/**
 * The randomwalk benchmark - PUSH variant
 */
template <class Graph, typename value_t>
class ll_b_randomwalk_push_ext : public ll_benchmark<Graph> {

	value_t d;
	int32_t max;
	int32_t path_len;

	value_t* G_pg_rank;


public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_push_ext(int32_t max, value_t d=0.85)
		: ll_benchmark<Graph>(
				/* assuming that value_t is either a float or a double */
				sizeof(value_t) == sizeof(float)
					? "_randomwalk_<float> - Push"
					: "randomwalk<double> - Push") {

        this->path_len = 150;
		assert(d > 0 && d < 1);
		this->d = d;
		this->max = max;
		this->create_auto_array_for_nodes(G_pg_rank);
	}


	/**
	 * Destroy the benchmark
	 */
	virtual ~ll_b_randomwalk_push_ext(void) {
	}


	/**
	 * Run the benchmark
	 *
	 * @return the numerical result, if applicable
	 */
	virtual double run(void) {
		printf("Random Walk\twalk_len<%d>, iteration<%d>\n", path_len, max);

		Graph& G = *this->_graph;
		srand(time(NULL));

#pragma omp parallel for
		for (node_t t0 = 0; t0 < G.max_nodes(); t0 ++)  {
			G.set_node_prop(G_pg_rank, t0, (value_t) 0.0); emulate_latency();
		}

#pragma omp parallel
        	{
#pragma omp for schedule(static, 1)
			for(int32_t iter=0; iter<max; iter++){
       	node_t src = (ll_rand64_positive()%(G.max_nodes()));

       	int t_degree = G.out_degree(src);
				if (t_degree == 0) continue;
				ATOMIC_ADD<value_t>(&G_pg_rank[src], 1.0);
				emulate_latency();

				for(int32_t k=1; k<path_len; k++){
					vector<node_t> out_deg;
					ll_foreach_out(w, G, src) { out_deg.push_back(w); }
          if(out_deg.empty()) { break; }
					src = out_deg[ll_rand64_positive() % out_deg.size()];
					ATOMIC_ADD<value_t>(&G_pg_rank[src], 1.0);
					emulate_latency();
				}
			}
		}

		return 0;
	}


	/**
	 * Finalize the benchmark
	 *
	 * @return the updated numerical result, if applicable
	 */
	virtual double finalize(void) {
		return 37.0; // dummy value
	}


	/**
	 * Print the results
	 *
	 * @param f the output file
	 */
	virtual void print_results(FILE* f) {
		print_results_part(f, this->_graph, G_pg_rank);
	}
};


/**
 * Adapters
 */


/**
 * The randomwalk benchmark - Pull variant, float
 */
template <class Graph>
class ll_b_randomwalk_pull_float
	: public ll_b_randomwalk_pull_ext<Graph, float> {

public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_pull_float(int32_t max, float d=0.85)
		: ll_b_randomwalk_pull_ext<Graph, float>(max, d) {}
};


/**
 * The randomwalk benchmark - Push variant, float
 */
template <class Graph>
class ll_b_randomwalk_push_float
	: public ll_b_randomwalk_push_ext<Graph, float> {

public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_push_float(int32_t max, float d=0.85)
		: ll_b_randomwalk_push_ext<Graph, float>(max, d) {}
};


/**
 * The randomwalk benchmark - Pull variant, double
 */
template <class Graph>
class ll_b_randomwalk_pull_double
	: public ll_b_randomwalk_pull_ext<Graph, double> {

public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_pull_double(int32_t max, double d=0.85)
		: ll_b_randomwalk_pull_ext<Graph, double>(max, d) {}
};


/**
 * The randomwalk benchmark - Push variant, double
 */
template <class Graph>
class ll_b_randomwalk_push_double
	: public ll_b_randomwalk_push_ext<Graph, double> {

public:

	/**
	 * Create the benchmark
	 */
	ll_b_randomwalk_push_double(int32_t max, double d=0.85)
		: ll_b_randomwalk_push_ext<Graph, double>(max, d) {}
};

#endif
