/*
 * ll_bfs_template.h
 * LLAMA Graph Analytics
 *
 * Copyright 2014
 *      The President and Fellows of Harvard College.
 *
 * Copyright 2014
 *      Oracle Labs.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS TERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This file was adapted from Green-Marl, which includes the following notice:
 *
 * Copyright (c) 2011-2012 Stanford University, unless otherwise specified.
 * All rights reserved.
 *
 * This software was developed by the Pervasive Parallelism Laboratory of
 * Stanford University, California, USA.
 *
 * Permission to use, copy, modify, and distribute this software in source or
 * binary form for any purpose with or without fee is hereby granted, provided
 * that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Stanford University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS TERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#ifndef LL_BFS_TEMPLATE_H
#define LL_BFS_TEMPLATE_H
#include <omp.h>
#include <string.h>
#include <map>
#include <set>
#include <unordered_set>
#include <unordered_map>

#include "mt_utils.h"
#include "nvgraph.h"

// single snapshot flag
#define FORCE_L0

// template<class Graph, typename T, bool use_multithread>
template<typename T, typename S>
class ll_bfs_template {

  public:
    ll_bfs_template(Graph<T,S>& _G, T _num_thread, T _snp_id) :
            G(_G), num_thread(_num_thread), snp_id(_snp_id)  {
        visited_bitmap = NULL; // bitmap
        visited_level = NULL;
        thread_local_next_level = NULL;
        down_edge_array = NULL;
        down_edge_set = NULL;
        down_edge_array_w = NULL;
    }

    virtual ~ll_bfs_template() {
        delete [] visited_bitmap;
        delete [] visited_level;
        delete [] thread_local_next_level;
        delete down_edge_set;

		if (down_edge_array != NULL) {
#ifndef FORCE_L0
			for (T i = 0; i < G.num_levels(); i++) delete[] down_edge_array[i];
#endif
			delete[] down_edge_array;
		}
    }

    void prepare(T root_node) {
		// TODO Is this correct? Do we need to poll a some sort of a runtime?
		prepare(root_node, num_thread);
	}

    void prepare(T root_node, T max_num_thread) {
        if (num_thread==1) {
            use_multithread = false;
        } else {
            use_multithread = true;
        }
		max_threads = num_thread;

        if(G.snptbl.empty()) {
            single_level = true;
            max_nodes = G.ntbl.size();
        } else {
            single_level = false;
            max_nodes = G.snptbl[snp_id].max_id + 1;
        }
        is_finished = false;
        curr_level = 0;
        root = root_node;
        state = ST_SMALL;
        // assert(root != LL_NIL_NODE);
        assert(root > 0);

        global_vector.clear();
        level_queue_begin.clear();
        level_count.clear();
        // create local queues
        if (thread_local_next_level == NULL) {
            thread_local_next_level = new std::vector<T>[num_thread];
            for (T i = 0; i < num_thread; i++)
                thread_local_next_level[i].reserve(THRESHOLD2);
        } else {
            for (T i = 0; i < num_thread; i++)
                thread_local_next_level[i].clear();
        }
    }

    void do_bfs_forward() {
        //---------------------------------
        // prepare root node
        //---------------------------------
        curr_level = 0;
        curr_count = 0;
        next_count = 0;

        small_visited[root] = curr_level;
        curr_count++;
        global_vector.push_back(root);
        global_curr_level_begin = 0;
        global_next_level_begin = curr_count;

        level_count.push_back(curr_count);
        level_queue_begin.push_back(global_curr_level_begin);

        bool is_done = false;
        while (!is_done) {
            switch (state) {
                case ST_SMALL: {
                    for (T i = 0; i < curr_count; i++) {
                        T t = global_vector[global_curr_level_begin + i];
                        iterate_neighbor_small(t);
                        visit_fw(t);            // visit after iteration. in that way, one can check  down-neighbors quite easily
                    }
                    break;
                }
                case ST_QUE: {
                    if (use_multithread)  // do it in parallel
                    {
                        T num_threads = std::min((T) max_threads, curr_count/128+1);
                        #pragma omp parallel num_threads(num_threads)
                        {
                            T tid = omp_get_thread_num();
                            #pragma omp for nowait
                            for (T i = 0; i < curr_count; i++) {
                                T t = global_vector[global_curr_level_begin + i];
                                iterate_neighbor_que(t, tid);
                                visit_fw(t);
                            }
                            finish_thread_que(tid);
                        }
                    }
                    else { // do it in sequential
                            T tid = 0;
                            for (T i = 0; i < curr_count; i++) {
                                //T t = global_curr_level[i];
                                T t = global_vector[global_curr_level_begin + i];
                                iterate_neighbor_que(t, tid);
                                visit_fw(t);
                            }
                            finish_thread_que(tid);
                    }
                    break;
                }
                case ST_Q2R: {
                    if (use_multithread) {  // do it in parallel
                        T num_threads = std::min((T) max_threads, curr_count/128+1);
                        #pragma omp parallel num_threads(num_threads)
                        {
                            T local_cnt = 0;
                            #pragma omp for nowait
                            for (T i = 0; i < curr_count; i++) {
                                T t = global_vector[global_curr_level_begin + i];
                                iterate_neighbor_rd(t, local_cnt);
                                visit_fw(t);
                            }
                            finish_thread_rd(local_cnt);
                        }
                    } else { // do it sequentially
                            T local_cnt = 0;
                            for (T i = 0; i < curr_count; i++) {
                                //T t = global_curr_level[i];
                                T t = global_vector[global_curr_level_begin + i];
                                iterate_neighbor_rd(t, local_cnt);
                                visit_fw(t);
                            }
                            finish_thread_rd(local_cnt);
                    }
                    break;
                }

                case ST_RD: {
                    if (use_multithread) { // do it in parallel
                        #pragma omp parallel
                        {
                            T local_cnt = 0;
                            #pragma omp for nowait schedule(dynamic,128)
                            for (T t = 0; t < max_nodes; t++) {
                                if (visited_level[t] == curr_level) {
                                    iterate_neighbor_rd(t, local_cnt);
                                    visit_fw(t);
                                }
                            }
                            finish_thread_rd(local_cnt);
                        }
                    } else { // do it in sequential
                            T local_cnt = 0;
                            for (T t = 0; t < max_nodes; t++) {
                                if (visited_level[t] == curr_level) {
                                    iterate_neighbor_rd(t, local_cnt);
                                    visit_fw(t);
                                }
                            }
                            finish_thread_rd(local_cnt);
                    }
                    break;
                }
                case ST_R2Q: {
                    if (use_multithread) { // do it in parallel
                        #pragma omp parallel
                        {
                            T tid = omp_get_thread_num();
                            #pragma omp for nowait schedule(dynamic,128)
                            for (T t = 0; t < max_nodes; t++) {
                                if (visited_level[t] == curr_level) {
                                    iterate_neighbor_que(t, tid);
                                    visit_fw(t);
                                }
                            }
                            finish_thread_que(tid);
                        }
                    } else {
                            T tid = 0;
                            for (T t = 0; t < max_nodes; t++) {
                                if (visited_level[t] == curr_level) {
                                    iterate_neighbor_que(t, tid);
                                    visit_fw(t);
                                }
                            }
                            finish_thread_que(tid);
                    }
                    break;
                }
            } // end of switch

            do_end_of_level_fw();
            is_done = get_next_state();

        } // end of while
    }

    void do_bfs_reverse() {
        // This function should be called only after do_bfs_foward has finished.
        // assumption: small-world graph
        T& level = curr_level;
        while (true) {
            T count = level_count[level];
            //T* queue_ptr = level_start_ptr[level];
            T* queue_ptr;
            T begin_idx = level_queue_begin[level];
            if (begin_idx == -1) {
                queue_ptr = NULL;
            } else {
                queue_ptr = & (global_vector[begin_idx]);
            }

            if (queue_ptr == NULL) {
#pragma omp parallel if (use_multithread)
                {
#pragma omp for nowait schedule(dynamic,128)
                    for (T i = 0; i < max_nodes; i++) {
                        if (visited_level[i] != curr_level) continue;
                        visit_rv(i);
                    }
                }
            } else {
				T num_threads = std::min((T) max_threads, curr_count/128+1);
#pragma omp parallel num_threads(num_threads) if (use_multithread)
                {
#pragma omp for nowait
                    for (T i = 0; i < count; i++) {
                        T u = queue_ptr[i];
                        visit_rv(u);
                    }
                }
            }

            do_end_of_level_rv();
            if (level == 0) break;
            level--;
        }
    }

//     bool is_down_edge(edge_t idx) {
//         if (state == ST_SMALL)
//             return (down_edge_set->find(idx) != down_edge_set->end());
//         else {
// #ifdef FORCE_L0
//             return down_edge_array[idx];
// #else
// 			T level = LL_EDGE_LEVEL(idx);
// 			if (level == LL_WRITABLE_LEVEL) {
// 				return down_edge_array_w[LL_EDGE_GET_WRITABLE(idx)->we_numerical_id];
// 			}
// 			return down_edge_array[level][LL_EDGE_INDEX(idx)];
// #endif
// 		}
//     }

  protected:
    virtual void visit_fw(T t)=0;
    virtual void visit_rv(T t)=0;
    // virtual bool check_navigator(T t, edge_t nx)=0;
    virtual void do_end_of_level_fw() {
    }
    virtual void do_end_of_level_rv() {
    }

    T get_root() {
        return root;
    }

    T get_level(T t) {
        // GCC expansion
        if (__builtin_expect((state == ST_SMALL), 0)) {
            if (small_visited.find(t) == small_visited.end())
                return __INVALID_LEVEL;
            else
                return small_visited[t];
        } else {
            return visited_level[t];
        }
    }

    T get_curr_level() {
        return curr_level;
    }


  private:
    bool get_next_state() {
        //const char* state_name[5] = {"SMALL","QUEUE","Q2R","RD","R2Q"};

        if (next_count == 0) return true;  // BFS is finished

        T next_state = state;
        switch (state) {
            case ST_SMALL:
                if (next_count >= THRESHOLD1) {
                    prepare_que();
                    next_state = ST_QUE;
                }
                break;
            case ST_QUE:
                if ((next_count >= THRESHOLD2) && (next_count >= curr_count*5)) {
                    prepare_read();
                    next_state = ST_Q2R;
                }
                break;
            case ST_Q2R:
                next_state = ST_RD;
                break;
            case ST_RD:
                if (next_count <= (2 * curr_count)) {
                    next_state = ST_R2Q;
                }
                break;
            case ST_R2Q:
                next_state = ST_QUE;
                break;
        }

        finish_level(state);
        state = next_state;

        return false;
    }

    void finish_level(T state) {
        if ((state == ST_RD) || (state == ST_Q2R)) {
            // output queue is not valid
        } else { // move output queue
            //T* temp = &(global_next_level[next_count]);
            //global_curr_level = global_next_level;
            //global_next_level = temp;
            global_curr_level_begin = global_next_level_begin;
            global_next_level_begin = global_next_level_begin + next_count;
        }

        curr_count = next_count;
        next_count = 0;
        curr_level++;

        // save 'new current' level status
        level_count.push_back(curr_count);
        if ((state == ST_RD) || (state == ST_Q2R)) {
            //level_start_ptr.push_back(NULL);
            level_queue_begin.push_back(-1);
        } else {
            //level_start_ptr.push_back(global_curr_level);
            level_queue_begin.push_back(global_curr_level_begin);
        }
    }

    void iterate_neighbor_small(T t) {

        // get list of out-deg neighbor
        vector<_prop<T,S>> out_degl;
       	get_out_degl(G, t, snp_id, out_degl);

        typename vector<_prop<T,S>>::iterator iter = out_degl.begin();
        for (; iter != out_degl.end(); iter++) {

            // check visited
            if (small_visited.find(iter->id) == small_visited.end()) {
                small_visited[iter->id] = curr_level + 1;
                //global_next_level[next_count++] = u;
                global_vector.push_back(iter->id);
                next_count++;
            }
        }
    }

    void prepare_que() {

        global_vector.reserve(max_nodes);

        // create bitmap and edges
        if (visited_bitmap == NULL) {
            visited_bitmap = new unsigned char[(max_nodes + 7) / 8];
            visited_level = new T[max_nodes];
        }

        if (use_multithread) {
			#pragma omp parallel
            {
				#pragma omp for nowait
                for (T i = 0; i < (max_nodes + 7) / 8; i++)
                    visited_bitmap[i] = 0;

				#pragma omp for nowait
                for (T i = 0; i < max_nodes; i++)
                    visited_level[i] = __INVALID_LEVEL;
            }
        } else {
            for (T i = 0; i < (max_nodes + 7) / 8; i++)
                visited_bitmap[i] = 0;
            for (T i = 0; i < max_nodes; i++)
                visited_level[i] = __INVALID_LEVEL;
        }

        //typename std::unordered_map<T, T>::iterator II;
        typename std::map<T, T>::iterator II;
        for (II = small_visited.begin(); II != small_visited.end(); II++) {
            T u = II->first;
            T lev = II->second;
            _ll_set_bit(visited_bitmap, u);
            visited_level[u] = lev;
        }
    }

    void iterate_neighbor_que(T t, T tid) {

        // get list of out-deg neighbor
        vector<_prop<T,S>> out_degl;
        get_out_degl(G, t, snp_id, out_degl);

        typename vector<_prop<T,S>>::iterator iter = out_degl.begin();
        for (; iter != out_degl.end(); iter++) {

            // assert(0 < iter->id && iter->id < max_nodes);
            if(iter->id <=0 || iter->id >= max_nodes){
              printf("assertion_fail: id<%u>, max_nodes<%u>\n", iter->id, max_nodes);
            }

            // check visited bitmap
            // test & test& set
            if (_ll_get_bit(visited_bitmap, iter->id) == 0) {
                bool re_check_result;
                if (use_multithread) {
                    re_check_result = _ll_set_bit_atomic(visited_bitmap, iter->id);
                } else {
                    re_check_result = true;
                    _ll_set_bit(visited_bitmap, iter->id);
                }

                if (re_check_result) {
                    // add to local q
                    thread_local_next_level[tid].push_back(iter->id);
                    visited_level[iter->id] = (curr_level + 1);
                }
            }
        }
    }

    void finish_thread_que(T tid) {
        T local_cnt = thread_local_next_level[tid].size();
        //copy curr_cnt to next_cnt
        if (local_cnt > 0) {
            T old_idx = __sync_fetch_and_add(&next_count, local_cnt);
            // copy to global vector
            memcpy(&(global_vector[global_next_level_begin + old_idx]),
                   &(thread_local_next_level[tid][0]),
                   local_cnt * sizeof(T));
        }
        thread_local_next_level[tid].clear();
    }

    void prepare_read() {
        // nothing to do
    }

    void iterate_neighbor_rd(T t, T& local_cnt) {

        // get list of out-deg neighbor
        vector<_prop<T,S>> out_degl;
        get_out_degl(G, t, snp_id, out_degl);

        typename vector<_prop<T,S>>::iterator iter = out_degl.begin();
        for (; iter != out_degl.end(); iter++) {

            // check visited bitmap
            // test & test& set
            if (_ll_get_bit(visited_bitmap, iter->id) == 0) {
                bool re_check_result;
                if (use_multithread) {
                    re_check_result = _ll_set_bit_atomic(visited_bitmap, iter->id);
                } else {
                    re_check_result = true;
                    _ll_set_bit(visited_bitmap, iter->id);
                }

                if (re_check_result) {
                    // add to local q
                    visited_level[iter->id] = curr_level + 1;
                    local_cnt++;
                }
            }
        }
    }

    void finish_thread_rd(T local_cnt) {
        __sync_fetch_and_add(&next_count, local_cnt);
    }

    //-----------------------------------------------------
    //-----------------------------------------------------
    static const T ST_SMALL = 0;
    static const T ST_QUE = 1;
    static const T ST_Q2R = 2;
    static const T ST_RD = 3;
    static const T ST_R2Q = 4;
    static const T THRESHOLD1 = 128;  // single threaded
    static const T THRESHOLD2 = 1024; // move to RD-based

    // not -1.
    //(why? because curr_level-1 might be -1, when curr_level = 0)
    static const T __INVALID_LEVEL = -2;

    T state;

    unsigned char* visited_bitmap; // bitmap
    T* visited_level; // assumption: small_world graph
    bool is_finished;
    T curr_level;
    T root;
    Graph<T,S>& G;
    T curr_count;
    T next_count;

    //std::unordered_map<T, T> small_visited;
    std::map<T, T> small_visited;
    // std::unordered_set<edge_t>* down_edge_set;
    std::unordered_set<T>* down_edge_set;
	  unsigned char* down_edge_array_w;
#ifdef FORCE_L0
    unsigned char* down_edge_array;
#else
    unsigned char** down_edge_array;
#endif

    //T* global_next_level;
    //T* global_curr_level;
    //T* global_queue;
    std::vector<T> global_vector;
    T global_curr_level_begin;
    T global_next_level_begin;

    //std::vector<T*> level_start_ptr;
    std::vector<T> level_queue_begin;
    std::vector<T> level_count;

    std::vector<T>* thread_local_next_level;
    bool use_multithread;
    T num_thread;
    bool single_level;
	  T max_threads;
    T snp_id;
    T max_nodes;
};

#endif
