/*
 * NVGraph :
 *
 *
 */

#ifndef NVGRAPH_H
#define NVGRAPH_H

/*************************** library ******************************************/
#include <iostream>
#include <fstream>
#include <vector>
#include <queue>
#include <string>
#include <sstream>
#include <map>
#include <set>
#include <dirent.h>
#include <omp.h>
#include <math.h>
#include <algorithm>
#include <assert.h>
#include <time.h>
#include <chrono>

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

/*************************** namespace ****************************************/
using namespace std;

/*************************** define *******************************************/
#define DRAM false
#define NVBM true
uint64_t nvram_update_counter = 0;
unsigned char* mem_dram;

#define db 10

/*************************** data structure ***********************************/
template<typename T, typename S>
struct xs0 {
  T id;
  S weight;
};

template<typename T>
struct edge_list {
  vector<T> dram;
  vector<T> nvbm;
  edge_list() {}
  edge_list(vector<T> e1, vector<T> e2) : dram(e1), nvbm(e2) {}
};

template<typename T>
struct node {
  short snapshot;  // snapshot id
  bool mem_type;    // true == nvbm ; false == dram
  T offset;  // index on nvbm or dram array
  T len;
  node(short a, bool b, T c, T d) :
      snapshot(a), mem_type(b), offset(c), len(d) {}
};

template<typename T>
struct _pair {
  T id;
  T offset;
  T len;
  _pair(T _id, T _offset, T _len) :
      id(_id), offset(_offset), len(_len) {}
};

template<typename T, typename S>
struct _prop {
  T id;
  S prop;
  _prop(T _id, S _prop) : id(_id), prop(_prop) {}
};

template<typename T, typename S>
struct snapshot {
  T max_id;
  vector<_pair<T>> modified;
  vector<_prop<T,S>> modified_neighbor;
  snapshot() {}
  snapshot(T id) : max_id(id) {}
  snapshot(vector<_pair<T>> b, vector<_prop<T,S>> c) :
           modified(b), modified_neighbor(c) {}
};

template<typename T>
struct property {
  string name;
  vector<T> prop_dram;
  vector<T> prop_nvbm;
  property() {}
  property(vector<T> p, vector<T> q) :
            prop_dram(p), prop_nvbm(q) {}
  property(string s, vector<T> p, vector<T> q) :
            name(s), prop_dram(p), prop_nvbm(q) {}
};

/* graph */
template<typename T, typename S>
struct Graph {
  short id = 0;
  vector<node<T>> ntbl;
  edge_list<T> etbl;
  property<S> wtbl;
  vector<snapshot<T,S>> snptbl;

  Graph() {}
  // snapshot graph
  Graph(vector<node<T>> a, edge_list<T> b, property<S> c) :
        ntbl(a), etbl(b), wtbl(c) {}
  // general graph
  Graph(vector<node<T>> a, edge_list<T> b, property<S> c,
        vector<snapshot<T,S>> d) : ntbl(a), etbl(b), wtbl(c), snptbl(d) {}
};

/* based graph */
template<typename T, typename S>
struct Graph_B {
  short id = 0;
  vector<node<T>> ntbl;
  edge_list<T> etbl;
  vector<snapshot<T,S>> snptbl;
  property<S> wtbl;
  Graph_B() {}
  Graph_B(vector<node<T>> a, edge_list<T> b, vector<snapshot<T,S>> d,
        property<S> e) : ntbl(a), etbl(b), snptbl(d), wtbl(e) {}
};

/* flatten */
template<typename T, typename S>
struct Graph_F {
  short id = 0;
  vector<node<T>> ntbl;
  edge_list<T> etbl;
  property<S> wtbl;
  Graph_F() {}
  Graph_F(vector<node<T>> a, edge_list<T> b, property<S> e) :
        ntbl(a), etbl(b), wtbl(e) {}
};

template<typename T, typename S>
struct element {
  T id; S property;
  element() {}
  element(T i, S p) : id(i), property(p) {}
};

template<typename T, typename S>
struct compare {
  bool operator() (const element<T,S>& e1, const element<T,S>& e2) const {
    return e1.id < e2.id;
  }

  bool operator() (const _prop<T,S>& e1, const _prop<T,S>& e2) const {
    return e1.id < e2.id;
  }
};

/*************************** nvram emulator ***********************************/
#define gettime asm_rdtsc
#define CPUFREQ 2200LLU /* GHz */
uint64_t idle_time_ns = 1000;

inline unsigned long long asm_rdtsc(void) {
    unsigned hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return ( (unsigned long long)lo)|( ((unsigned long long)hi)<<32 );
}

inline void emulate_latency_cycles(uint64_t cycles) {
    uint64_t start;
    uint64_t stop;

    start = asm_rdtsc();

    do {
        /* RDTSC doesn't necessarily wait for previous instructions to complete
         * so a serializing instruction is usually used to ensure previous
         * instructions have completed. However, in our case this is a desirable
         * property since we want to overlap the latency we emulate with the
         * actual latency of the emulated instruction.
         */
        stop = asm_rdtsc();
    } while (stop - start < cycles);
}

inline void emulate_latency_cycles_ns(void) {
  uint64_t cycles = idle_time_ns * CPUFREQ / 1000;
  emulate_latency_cycles(cycles);
}

inline void emulate_latency(void) {
  emulate_latency_cycles_ns();
}

inline void nvram_cnt(void) { nvram_update_counter += 1; }

/*************************** function declaration *****************************/
inline unsigned long long asm_rdtsc(void);
inline void emulate_latency_cycles(int cycles);
inline void emulate_latency_cycles_ns(void);
inline void emulate_latency(void);

template<typename T>
inline T get_current_time(void);

template<typename T>
inline void show_runtime(string mseg, const T timestamp);

template<typename T>
inline T timeval_to_ms(struct timeval t);

template<typename T, typename S>
T calculate_default_sz(const char* fname);

template<typename T, typename S>
bool get_next(FILE* fd, xs0<T,S>* out);

void load_mem_type(const char* fname, unsigned char* mem_dram);

template<typename T, typename S>
T load_based_graph(const char* fname, const unsigned char*  mem_dram,
  vector<node<T>>& ntbl, property<S>& wtbl, edge_list<T>& etbl);

template<typename T, typename S>
bool load_delta_graph(const char* fname, vector<snapshot<T,S>>& snptbl);

template<typename T, typename S>
bool extend_based_graph(const unsigned char* mem_dram, const vector<snapshot<T,S>>& snptbl,
  vector<node<T>>& ntbl, property<S>& wtbl, edge_list<T>& etbl);

template<typename T, typename S>
bool inline is_exist(const vector<snapshot<T,S>>& snptbl, const short snp_id, const T n);

template<typename T, typename S>
void print_graph(const Graph_F<T,S>& g);

template<typename T, typename S>
void print_graph(const Graph_B<T,S>& g);

template<typename T, typename S>
void print_edge_table_helper(const vector<T>& etbl, const vector<S>& wtbl);

template<typename T, typename S>
void print_edge_table(const edge_list<T>& tbl, const property<S>& wtbl);

template<typename T>
void print_node_table(const vector<node<T>>& tbl);

template<typename T, typename S>
void print_snapshot(const vector<snapshot<T,S>>& snptbl);

template<typename T, typename S>
T get_max_id(const Graph<T,S>& g);

template<typename T, typename S>
T get_num_out_deg(const Graph<T,S>& g, const T n);

template<typename T, typename S>
T get_in_degree(const Graph<T,S>& g, const T n);

template<typename T, typename S>
T get_rand_out_deg(const Graph<T,S>& g, const T n);

template<typename T, typename S>
void get_out_degl(const Graph<T,S>& g, const T n, T& beg, T& end);

template<typename T, typename S>
void flatten(const Graph<T,S>& G, Graph<T,S>& g, const T snp_id);

template<typename T>
void get_filelist(char* directory, vector<string>& files);

template<typename T, typename S>
void graph_persist(const Graph_B<T,S>& g);

#endif
